# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.4
dpt_vfs/dpt_vfs/file/watcher_pyinotify.py
"""

# pylint: disable=import-error,no-name-in-module,unused-argument

from weakref import ref
from os import path

from dpt_logging import LogLine
from dpt_module_loader import NamedClassLoader
from dpt_threading import InstanceLock, ThreadLock
import pyinotify

try: from pyinotify import IN_ATTRIB, IN_CLOSE_WRITE, IN_CREATE, IN_DELETE, IN_DELETE_SELF, IN_MODIFY, IN_MOVE_SELF, IN_MOVED_FROM, IN_MOVED_TO
except ImportError: from pyinotify.EventsCodes import IN_ATTRIB, IN_CLOSE_WRITE, IN_CREATE, IN_DELETE, IN_DELETE_SELF, IN_MODIFY, IN_MOVE_SELF, IN_MOVED_FROM, IN_MOVED_TO

from .watcher_pyinotify_callback import WatcherPyinotifyCallback

class WatcherPyinotify(pyinotify.WatchManager):
    """
"file:///" watcher using pyinotify's ThreadedNotifier.

:author:     direct Netware Group et al.
:copyright:  direct Netware Group - All rights reserved
:package:    dpt
:subpackage: vfs
:since:      v1.0.0
:license:    https://www.direct-netware.de/redirect?licenses;mpl2
             Mozilla Public License, v. 2.0
    """

    # pylint: disable=bad-option-value,slots-on-old-class
    __slots__ = [ "__weakref__",
                  "_lock",
                  "_pyinotify_instance",
                  "watched_callbacks",
                  "watched_path_files",
                  "watched_paths"
                ]
    """
python.org: __slots__ reserves space for the declared variables and prevents
the automatic creation of __dict__ and __weakref__ for each instance.
    """
    _instance_lock = InstanceLock()
    """
Thread safety instance lock
    """
    _weakref_instance = None
    """
WatcherPyinotify weakref instance
    """

    def __init__(self):
        """
Constructor __init__(WatcherPyinotify)

:since: v1.0.0
        """

        pyinotify.WatchManager.__init__(self)

        self._lock = ThreadLock()
        """
    Thread safety lock
        """
        self._pyinotify_instance = None
        """
pyinotify instance
        """
        self.watched_callbacks = { }
        """
Callbacks for watched files
        """
        self.watched_path_files = { }
        """
pyinotify watch fds
        """
        self.watched_paths = { }
        """
pyinotify watch fds
        """

        self._init_notifier()
    #

    def __del__(self):
        """
Destructor __del__(WatcherPyinotify)

:since: v1.0.0
        """

        self.stop()
    #

    def check(self, _path):
        """
Checks a given path for changes if "is_synchronous()" is true.

:param _path: Filesystem path

:return: (bool) True if the given path URL has been changed since last check
         and "is_synchronous()" is true.
:since:  v1.0.0
        """

        return False
    #

    def free(self):
        """
Frees all watcher callbacks for garbage collection.

:since: v1.0.0
        """

        with self._lock:
            if (len(self.watched_paths) > 0):
                self.watched_callbacks = { }
                self.watched_paths = { }
            #
        #
    #

    def _init_notifier(self):
        """
Initializes the pyinotify instance.

:since: v1.0.0
        """

        with self._lock:
            if (self._pyinotify_instance is None):
                LogLine.debug("{0!r} mode is asynchronous", self, context = "dpt_vfs")

                self._pyinotify_instance = pyinotify.ThreadedNotifier(self, WatcherPyinotifyCallback(self), timeout = 5000)
                self._pyinotify_instance.start()
            #
        #
    #

    def is_watched(self, _path, callback = None):
        """
Returns true if the filesystem path is already watched. It will return false
if a callback is given but not defined for the watched path.

:param _path: Filesystem path
:param callback: Callback to be checked for the watched filesystem path

:return: (bool) True if watched with the defined callback or any if not
         defined.
:since:  v1.0.0
        """

        _return = False

        with self._lock:
            if (_path in self.watched_callbacks): _return = (True if (callback is None) else (callback in self.watched_callbacks[_path]))
        #

        return _return
    #

    def get_callbacks(self, _path):
        """
Returns all registered callbacks for the given path.

:param _path: Filesystem path

:return: (list) List of watcher callbacks
:since:  v1.0.0
        """

        _return = [ ]

        with self._lock:
            if (_path in self.watched_callbacks): _return = self.watched_callbacks[_path]
            elif (not path.isdir(_path)): _return = self.get_callbacks(path.split(_path)[0])
        #

        return _return
    #

    def register(self, _path, callback):
        """
Handles registration of filesystem watches and its callbacks.

:param _path: Filesystem path to be watched
:param callback: Callback for the path

:return: (bool) True on success
:since:  v1.0.0
        """

        # pylint: disable=no-member

        _return = True

        with self._lock:
            if (path.isdir(_path)): directory_path = _path
            else: directory_path = path.split(_path)[0]

            if (directory_path not in self.watched_paths):
                inotify_result = self.add_watch(directory_path, (IN_ATTRIB | IN_CLOSE_WRITE | IN_CREATE | IN_DELETE | IN_DELETE_SELF | IN_MODIFY | IN_MOVE_SELF | IN_MOVED_FROM | IN_MOVED_TO))

                if (inotify_result[directory_path] < 0): _return = False
                else: self.watched_paths.update(inotify_result)
            #

            if (_return):
                if (directory_path not in self.watched_path_files): self.watched_path_files[directory_path] = [ _path ]
                elif (_path not in self.watched_path_files[directory_path]): self.watched_path_files[directory_path].append(_path)

                if (_path not in self.watched_callbacks): self.watched_callbacks[_path] = [ callback ]
                elif (callback not in self.watched_callbacks[_path]): self.watched_callbacks[_path].append(callback)
            #
        #

        return _return
    #

    def stop(self):
        """
Stops all watchers.

:since: v1.0.0
        """

        with self._lock:
            self.free()

            if (self._pyinotify_instance is not None):
                self._pyinotify_instance.stop()
                self._pyinotify_instance = None
            #
        #
    #

    def unregister(self, _path, callback, _deleted = False):
        """
Handles deregistration of filesystem watches.

:param _path: Filesystem path watched
:param callback: Callback for the path
:param _deleted: File has been deleted

:return: (bool) True on success
:since:  v1.0.0
        """

        # pylint: disable=no-member

        _return = True

        with self._lock:
            is_directory = path.isdir(_path)

            if (is_directory): directory_path = _path
            else: directory_path = path.split(_path)[0]

            if (directory_path in self.watched_path_files and _path in self.watched_paths):
                if (callback is None or _deleted): self.watched_callbacks[_path] = [ ]
                elif (callback in self.watched_callbacks[_path]): self.watched_callbacks[_path].remove(callback)

                if (len(self.watched_callbacks[_path]) < 1): del(self.watched_callbacks[_path])

                if (is_directory and _deleted):
                    for filepath in self.watched_path_files[directory_path]: self.unregister(filepath, None, True)
                #

                if (len(self.watched_path_files[directory_path]) < 1):
                    if (not _deleted): self.rm_watch(self.watched_paths[directory_path])

                    del(self.watched_path_files[directory_path])
                    del(self.watched_paths[directory_path])
                #
            else: _return = False
        #

        return _return
    #

    @classmethod
    def get_singleton(cls):
        """
Get the WatcherPyinotify singleton.

:param cls: Python class

:return: (object) Object on success
:since:  v1.0.0
        """

        # pylint: disable=not-callable

        _return = None

        with WatcherPyinotify._instance_lock:
            if (WatcherPyinotify._weakref_instance is None):
                log_handler = NamedClassLoader.get_singleton("dpt_logging.LogHandler", False)
                if (log_handler is not None): log_handler.add_logger("pyinotify")
            else: _return = WatcherPyinotify._weakref_instance()

            if (_return is None):
                _return = cls()
                WatcherPyinotify._weakref_instance = ref(_return)
            #
        #

        return _return
    #
#



# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.4
dpt_vfs/dpt_vfs/file/object.py
"""

# pylint: disable=import-error,invalid-name,no-member,no-name-in-module

from os import path
import os

try: from urllib.parse import quote_plus, unquote_plus
except ImportError: from urllib import quote_plus, unquote_plus

from dpt_file import File
from dpt_logging import LogLine
from dpt_mime_type import MimeType
from dpt_runtime.exceptions import IOException, OperationNotSupportedException

from ...abstract import Abstract
from ...file_like_wrapper_mixin import FileLikeWrapperMixin

if (hasattr(os, "PathLike")): _PathLike = os.PathLike
else:
    class _PathLike(object): pass
#

class Object(FileLikeWrapperMixin, _PathLike, Abstract):
    """
Provides the VFS implementation for 'file' objects.

:author:     direct Netware Group et al.
:copyright:  direct Netware Group - All rights reserved
:package:    dpt
:subpackage: vfs
:since:      v1.0.0
:license:    https://www.direct-netware.de/redirect?licenses;mpl2
             Mozilla Public License, v. 2.0
    """

    # pylint: disable=unused-argument

    _FILE_WRAPPED_METHODS = ( "flush",
                              "is_eof",
                              "read",
                              "seek",
                              "tell",
                              "truncate",
                              "write"
                            )
    """
File IO methods implemented by an wrapped resource.
    """

    __slots__ = [ "dir_path_name", "file_path_name", "object_readonly" ] + FileLikeWrapperMixin._mixin_slots_
    """
python.org: __slots__ reserves space for the declared variables and prevents
the automatic creation of __dict__ and __weakref__ for each instance.
    """

    def __init__(self):
        """
Constructor __init__(Object)

:since: v1.0.0
        """

        Abstract.__init__(self)
        FileLikeWrapperMixin.__init__(self)

        self.dir_path_name = None
        """
Directory path and name set for "TYPE_DIRECTORY"
        """
        self.file_path_name = None
        """
File path and name set for "TYPE_FILE"
        """
        self.object_readonly = None
        """
True to open the object and nested ones read-only
        """

        self.supported_features['filesystem_path_name'] = True
        self.supported_features['flush'] = self._supports_flush
        self.supported_features['implementing_instance'] = self._supports_implementing_instance
        self.supported_features['seek'] = self._supports_seek
        self.supported_features['time_created'] = True
        self.supported_features['time_updated'] = True
    #

    def __fspath__(self):
        """
python.org: Return the file system path representation of the object.

:return: (str) File system path representation
:since:  v1.0.0
        """

        return self.filesystem_path_name
    #

    @property
    def filesystem_path_name(self):
        """
Returns the path and name for the VFS object in the system filesystem.

:return: (str) System filesystem path and name of the VFS object
:since:  v1.0.0
        """

        _return = None

        if (self.dir_path_name is not None): _return = self.dir_path_name
        elif (self.file_path_name is not None): _return = self.file_path_name
        else: raise IOException("VFS object not opened")

        return _return
    #

    @property
    def implementing_instance(self):
        """
Returns the implementing instance.

:return: (mixed) Implementing instance
:since:  v1.0.0
        """

        _return = None

        if (self._wrapped_resource is None
            and self.file_path_name is not None
            ): self._open_wrapped_resource()

        if (self._wrapped_resource is not None): _return = self._wrapped_resource
        elif (self.dir_path_name is None): raise IOException("VFS object not opened")

        return _return
    #

    @property
    def implementing_scheme(self):
        """
Returns the implementing scheme name.

:return: (str) Implementing scheme name
:since:  v1.0.0
        """

        return "file"
    #

    @property
    def is_valid(self):
        """
Returns true if the object is available.

:return: (bool) True on success
:since:  v1.0.0
        """

        _return = False

        if (self._wrapped_resource is not None): _return = self._wrapped_resource.is_valid
        elif (self.dir_path_name is not None): _return = os.access(self.dir_path_name, os.X_OK)
        elif (self.file_path_name is not None): _return = os.access(self.file_path_name, os.R_OK)

        return _return
    #

    @property
    def mimetype(self):
        """
Returns the mime type of this VFS object.

:return: (str) VFS object mime type
:since:  v1.0.0
        """

        _return = None

        if (self.dir_path_name is not None): _return = "text/directory"
        elif (self.file_path_name is not None):
            file_data = path.splitext(self.file_path_name)
            mimetype_definition = MimeType.get_instance().get(file_data[1][1:])

            _return = ("application/octet-stream" if (mimetype_definition is None) else mimetype_definition['type'])
        else: raise IOException("VFS object not opened")

        return _return
    #

    @property
    def name(self):
        """
Returns the name of this VFS object.

:return: (str) VFS object name
:since:  v1.0.0
        """

        _return = None

        if (self.dir_path_name is not None): _return = path.basename(self.dir_path_name)
        elif (self.file_path_name is not None): _return = path.basename(self.file_path_name)
        else: raise IOException("VFS object not opened")

        return _return
    #

    @property
    def size(self):
        """
Returns the size in bytes.

:return: (int) Size in bytes
:since:  v1.0.0
        """

        _return = None

        if (self.dir_path_name is not None): _return = 0
        elif (self.file_path_name is not None): _return = os.stat(self.file_path_name).st_size
        else: raise IOException("VFS object not opened")

        return _return
    #

    @property
    def time_created(self):
        """
Returns the UNIX timestamp this object was created.

:return: (int) UNIX timestamp this object was created
:since:  v1.0.0
        """

        _return = None

        if (self.dir_path_name is not None): _return = os.stat(self.dir_path_name).st_ctime
        elif (self.file_path_name is not None): _return = os.stat(self.file_path_name).st_ctime
        else: raise IOException("VFS object not opened")

        return _return
    #

    @property
    def time_updated(self):
        """
Returns the UNIX timestamp this object was updated.

:return: (int) UNIX timestamp this object was updated
:since:  v1.0.0
        """

        _return = None

        if (self.dir_path_name is not None): _return = os.stat(self.dir_path_name).st_mtime
        elif (self.file_path_name is not None): _return = os.stat(self.file_path_name).st_mtime
        else: raise IOException("VFS object not opened")

        return _return
    #

    @property
    def type(self):
        """
Returns the type of this object.

:return: (int) Object type
:since:  v1.0.0
        """

        _return = None

        if (self.dir_path_name is not None): _return = Object.TYPE_DIRECTORY
        elif (self.file_path_name is not None): _return = Object.TYPE_FILE
        else: raise IOException("VFS object not opened")

        return _return
    #

    @property
    def url(self):
        """
Returns the URL of this VFS object.

:return: (str) VFS URL
:since:  v1.0.0
        """

        object_id = None

        if (self.dir_path_name is not None): object_id = quote_plus(self.dir_path_name, "/")
        elif (self.file_path_name is not None): object_id = quote_plus(self.file_path_name, "/")

        if (object_id is None): raise IOException("VFS object not opened")

        return "file:///{0}".format(object_id)
    #

    def close(self):
        """
python.org: Flush and close this stream.

:since: v1.0.0
        """

        if (self.dir_path_name is not None): self.dir_path_name = None
        else:
            try: FileLikeWrapperMixin.close(self)
            finally: self.file_path_name = None
        #
    #

    def _ensure_directory_readable(self, vfs_url, dir_path_name):
        """
Ensures that the given directory path readable.

:param vfs_url: VFS URL
:param dir_path_name: Directory path and name

:since: v1.0.0
        """

        if (not os.access(dir_path_name, os.X_OK)): raise IOException("VFS URL '{0}' is invalid".format(vfs_url))
    #

    def _ensure_directory_writable(self, vfs_url, dir_path_name):
        """
Ensures that the given directory path writable.

:param vfs_url: VFS URL
:param dir_path_name: Directory path and name

:since: v1.0.0
        """

        if (not os.access(dir_path_name, os.X_OK)): raise IOException("VFS URL '{0}' is invalid".format(vfs_url))
    #

    def new(self, _type, vfs_url):
        """
Creates a new VFS object.

:param _type: VFS object type
:param vfs_url: VFS URL

:since: v1.0.0
        """

        if (_type == Object.TYPE_DIRECTORY): self._new_directory(vfs_url)
        elif (_type == Object.TYPE_FILE): self._new_file(vfs_url)
        else: raise OperationNotSupportedException()
    #

    def _new_file(self, vfs_url):
        """
Creates a new VFS file object.

:param vfs_url: VFS URL

:since: v1.0.0
        """

        file_path_name = unquote_plus(Abstract._get_id_from_vfs_url(vfs_url))
        self._ensure_directory_writable(vfs_url, path.dirname(file_path_name))

        self._open_file(vfs_url, file_path_name)
    #

    def open(self, vfs_url, readonly = False):
        """
Opens a VFS object. The handle is set at the beginning of the object.

:param vfs_url: VFS URL
:param readonly: Open object in readonly mode

:since: v1.0.0
        """

        if (self.dir_path_name is not None
            or self.file_path_name is not None
           ): raise IOException("Can't create new VFS object on already opened instance")

        object_path_name = unquote_plus(Abstract._get_id_from_vfs_url(vfs_url))

        if (path.isdir(object_path_name)): self._open_directory(vfs_url, object_path_name, readonly)
        else: self._open_file(vfs_url, object_path_name, readonly)
    #

    def _open_directory(self, vfs_url, dir_path_name, readonly = False):
        """
Opens a VFS directory object.

:param vfs_url: VFS URL
:param dir_path_name: Directory path and name

:since: v1.0.0
        """

        self._ensure_directory_readable(vfs_url, dir_path_name)

        self.dir_path_name = path.abspath(dir_path_name)
        self.object_readonly = readonly
    #

    def _open_file(self, vfs_url, file_path_name, readonly = True):
        """
Opens (and creates) a VFS file object.

:param vfs_url: VFS URL
:param file_path_name: File path and name
:param readonly: Open file in readonly mode

:since: v1.0.0
        """

        self.file_path_name = path.abspath(file_path_name)
        self.object_readonly = readonly
    #

    def _open_wrapped_resource(self):
        """
Opens the wrapped resource once needed for an file IO request.

:since: v1.0.0
        """

        if (self.file_path_name is None): raise IOException("VFS object not opened")

        file_mode = ("rb" if (self.object_readonly) else "r+b")

        _file = File()
        if (_file.open(self.file_path_name, self.object_readonly, file_mode)): self._set_wrapped_resource(_file)
    #

    def scan(self):
        """
Scan over objects of a collection like a directory.

:return: (list) Child VFS objects
:since:  v1.0.0
        """

        if (self.file_path_name is not None): raise OperationNotSupportedException("VFS object can not be scanned")
        if (self.dir_path_name is None): raise IOException("VFS object not opened")

        _return = [ ]

        entry_list = os.listdir(self.dir_path_name)
        entry_list.sort()

        dir_path_url = self.url

        for entry in entry_list:
            if (entry[0] != "."):
                vfs_child_object = Object()

                try:
                    vfs_child_object.open("{0}/{1}".format(dir_path_url, entry), self.object_readonly)
                    _return.append(vfs_child_object)
                except IOException as handled_exception: LogLine.error(handled_exception, context = "dpt_vfs")
            #
        #

        return _return
    #

    def _supports_flush(self):
        """
Returns false if flushing buffers is not supported.

:return: (bool) True if flushing buffers is supported
:since:  v1.0.0
        """

        return (self.file_path_name is not None and (not self.object_readonly))
    #

    def _supports_implementing_instance(self):
        """
Returns false if no underlying, implementing instance can be returned.

:return: (bool) True if an implementing instance can be returned.
:since:  v1.0.0
        """

        return (self._wrapped_resource is not None)
    #

    def _supports_seek(self):
        """
Returns false if seek is not supported.

:return: (bool) True if seek is supported
:since:  v1.0.0
        """

        return (self.file_path_name is not None)
    #
#



# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.4
dpt_vfs/dpt_vfs/file/watcher_mtime.py
"""

# pylint: disable=import-error,no-name-in-module

import os

try: from urllib.parse import quote_plus
except ImportError: from urllib import quote_plus

from dpt_logging import ExceptionLogTrap
from dpt_threading import ThreadLock

from ...abstract_watcher import AbstractWatcher

class WatcherMtime(object):
    """
"file:///" watcher using os.stat to detect changes.

:author:     direct Netware Group et al.
:copyright:  direct Netware Group - All rights reserved
:package:    dpt
:subpackage: vfs
:since:      v1.0.0
:license:    https://www.direct-netware.de/redirect?licenses;mpl2
             Mozilla Public License, v. 2.0
    """

    __slots__ = [ ]
    """
python.org: __slots__ reserves space for the declared variables and prevents
the automatic creation of __dict__ and __weakref__ for each instance.
    """
    _lock = ThreadLock()
    """
Thread safety lock
    """
    _watched_callbacks = { }
    """
Callbacks for watched files
    """
    _watched_paths = { }
    """
Dict with latest modified timestamps
    """

    def check(self, _path):
        """
Checks a given path for changes if "is_synchronous()" is true.

:param _path: Filesystem path

:return: (bool) True if the given path URL has been changed since last check
         and "is_synchronous()" is true.
:since:  v1.0.0
        """

        _return = False

        event_type = None

        with WatcherMtime._lock:
            if (WatcherMtime._watched_paths is not None
                and _path in WatcherMtime._watched_paths
               ):
                modified_time = (os.stat(_path).st_mtime if (os.access(_path, os.R_OK)) else -1)

                if (modified_time < 0):
                    event_type = AbstractWatcher.EVENT_TYPE_DELETED
                    self.unregister(_path, None)
                elif (WatcherMtime._watched_paths[_path] != modified_time):
                    event_type = (AbstractWatcher.EVENT_TYPE_CREATED
                                  if (WatcherMtime._watched_paths[_path] < 0) else
                                  AbstractWatcher.EVENT_TYPE_MODIFIED
                                 )

                    WatcherMtime._watched_paths[_path] = modified_time
                #
            #
        #

        if (event_type is not None):
            _return = True

            url = "file:///{0}".format(quote_plus(_path, "/"))

            for callback in WatcherMtime._watched_callbacks[_path]:
                with ExceptionLogTrap("dpt_vfs"): callback(event_type, url)
            #
        #

        return _return
    #

    def free(self):
        """
Frees all watcher callbacks for garbage collection.

:since: v1.0.0
        """

        with WatcherMtime._lock:
            if (WatcherMtime._watched_paths is not None and len(WatcherMtime._watched_paths) > 0):
                WatcherMtime._watched_callbacks = None
                WatcherMtime._watched_paths = None
            #
        #
    #

    def is_watched(self, _path, callback = None):
        """
Returns true if the filesystem path is already watched. It will return false
if a callback is given but not defined for the watched path.

:param _path: Filesystem path
:param callback: Callback to be checked for the watched filesystem path

:return: (bool) True if watched with the defined callback or any if not
         defined.
:since:  v1.0.0
        """

        with WatcherMtime._lock:
            _return = (WatcherMtime._watched_paths is not None and _path in WatcherMtime._watched_paths)
            if (_return and callback is not None): _return = (callback in WatcherMtime._watched_callbacks[_path])
        #

        return _return
    #

    def register(self, _path, callback):
        """
Handles registration of filesystem watches and its callbacks.

:param _path: Filesystem path to be watched
:param callback: Callback for the path

:return: (bool) True on success
:since:  v1.0.0
        """

        _return = True

        with WatcherMtime._lock:
            if (WatcherMtime._watched_callbacks is not None):
                if (_path not in WatcherMtime._watched_paths):
                    WatcherMtime._watched_paths[_path] = (os.stat(_path).st_mtime if (os.access(_path, os.R_OK)) else -1)
                    WatcherMtime._watched_callbacks[_path] = [ ]
                #

                if (callback not in WatcherMtime._watched_callbacks[_path]): WatcherMtime._watched_callbacks[_path].append(callback)
            #
        #

        return _return
    #

    def stop(self):
        """
Stops all watchers.

:since: v1.0.0
        """

        self.free()
    #

    def unregister(self, _path, callback):
        """
Handles deregistration of filesystem watches.

:param _path: Filesystem path watched
:param callback: Callback for the path

:return: (bool) True on success
:since:  v1.0.0
        """

        _return = True

        with WatcherMtime._lock:
            if (WatcherMtime._watched_paths is not None and _path in WatcherMtime._watched_paths):
                if (callback is None): WatcherMtime._watched_callbacks[_path] = [ ]
                elif (callback in WatcherMtime._watched_callbacks[_path]):
                    WatcherMtime._watched_callbacks[_path].remove(callback)
                #

                if (len(WatcherMtime._watched_callbacks[_path]) < 1):
                    del(WatcherMtime._watched_callbacks[_path])
                    del(WatcherMtime._watched_paths[_path])
                #
            else: _return = False
        #

        return _return
    #
#



# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.4
dpt_vfs/abstract_watcher.py
"""

# pylint: disable=unused-argument

from dpt_runtime.exceptions import NotImplementedException

class AbstractWatcher(object):
    """
Abstract watcher for change events.

:author:     direct Netware Group et al.
:copyright:  direct Netware Group - All rights reserved
:package:    dpt
:subpackage: vfs
:since:      v1.0.0
:license:    https://www.direct-netware.de/redirect?licenses;mpl2
             Mozilla Public License, v. 2.0
    """

    EVENT_TYPE_CREATED = 1
    """
Created event
    """
    EVENT_TYPE_DELETED = 2
    """
Deleted event
    """
    EVENT_TYPE_MODIFIED = 3
    """
Created event
    """

    __slots__ = [ "__weakref__" ]
    """
python.org: __slots__ reserves space for the declared variables and prevents
the automatic creation of __dict__ and __weakref__ for each instance.
    """

    @property
    def implementing_scheme(self):
        """
Returns the implementing scheme name.

:return: (str) Implementing scheme name
:since:  v1.0.0
        """

        raise NotImplementedException()
    #

    @property
    def is_synchronous(self):
        """
Returns true if changes are only detected after "check()" has been
called.

:return: (bool) True if changes are not detected automatically
:since:  v1.0.0
        """

        return True
    #

    def check(self, url):
        """
Checks a given URL for changes if "is_synchronous()" is true.

:param url: Resource URL

:since: v1.0.0
        """

        raise NotImplementedException()
    #

    def free(self):
        """
Frees all watcher callbacks for garbage collection.

:since: v1.0.0
        """

        raise NotImplementedException()
    #

    def is_watched(self, url, callback = None):
        """
Returns true if the resource URL is already watched. It will return false
if a callback is given but not defined for the watched URL.

:param url: Resource URL
:param callback: Callback to be checked for the watched resource URL

:return: (bool) True if watched with the defined callback or any if not
         defined.
:since:  v1.0.0
        """

        raise NotImplementedException()
    #

    def register(self, url, callback):
        """
Handles registration of resource URL watches and its callbacks.

:param url: Resource URL to be watched
:param callback: Callback for the path

:return: (bool) True on success
:since:  v1.0.0
        """

        raise NotImplementedException()
    #

    def stop(self):
        """
Stops all watchers.

:since: v1.0.0
        """

        pass
    #

    def unregister(self, url, callback):
        """
Handles deregistration of resource URL watches.

:param url: Resource URL watched
:param callback: Callback for the path

:return: (bool) True on success
:since:  v1.0.0
        """

        raise NotImplementedException()
    #

    @staticmethod
    def disable():
        """
Disables this watcher and frees all callbacks for garbage collection.

:since: v1.0.0
        """

        raise NotImplementedException()
    #
#



# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.4
dpt_vfs/watcher_implementation.py
"""

from dpt_module_loader import NamedClassLoader
from dpt_runtime import Binary
from dpt_runtime.exceptions import IOException, OperationNotSupportedException, ValueException
from dpt_threading import ThreadLock

from .abstract_watcher import AbstractWatcher

class WatcherImplementation(object):
    """
"WatcherImplementation" provides implementation independent methods to
access VFS watchers.

:author:     direct Netware Group et al.
:copyright:  direct Netware Group - All rights reserved
:package:    dpt
:subpackage: vfs
:since:      v1.0.0
:license:    https://www.direct-netware.de/redirect?licenses;mpl2
             Mozilla Public License, v. 2.0
    """

    EVENT_TYPE_CREATED = AbstractWatcher.EVENT_TYPE_CREATED
    """
Created event
    """
    EVENT_TYPE_DELETED = AbstractWatcher.EVENT_TYPE_DELETED
    """
Deleted event
    """
    EVENT_TYPE_MODIFIED = AbstractWatcher.EVENT_TYPE_MODIFIED
    """
Created event
    """

    __slots__ = [ ]
    """
python.org: __slots__ reserves space for the declared variables and prevents
the automatic creation of __dict__ and __weakref__ for each instance.
    """
    _classes = { }
    """
True if existing watchers have been disabled and no new ones can be
initialized.
    """
    _is_disabled = False
    """
True if existing watchers have been disabled and no new ones can be
initialized.
    """
    _lock = ThreadLock()
    """
Thread safety lock
    """

    @staticmethod
    def disable():
        """
Disables all watchers and free callbacks for garbage collection.

:since: v1.0.0
        """

        with WatcherImplementation._lock:
            if (not WatcherImplementation._is_disabled):
                WatcherImplementation._is_disabled = True

                for vfs_scheme in WatcherImplementation._classes: WatcherImplementation._classes[vfs_scheme].disable()
                WatcherImplementation._classes = None
            #
        #
    #

    @staticmethod
    def get_class(scheme):
        """
Returns an VFS watcher class for the given scheme.

:return: (object) VFS watcher class
:since:  v1.0.0
        """

        with WatcherImplementation._lock:
            if (WatcherImplementation._is_disabled): raise OperationNotSupportedException("VFS watcher creation has been disabled")

            _return = WatcherImplementation._classes.get(scheme)

            if (_return is None):
                _return = NamedClassLoader.get_class_in_namespace("dpt_vfs", "{0}.Watcher".format(scheme.replace("-", "_")))

                if (issubclass(_return, AbstractWatcher)): WatcherImplementation._classes[scheme] = _return
                else: _return = None
            #
        #

        if (_return is None): raise IOException("VFS watcher not defined for URL scheme '{0}'".format(scheme))

        return _return
    #

    @staticmethod
    def get_instance(scheme):
        """
Returns an VFS watcher instance for the given scheme.

:return: (object) VFS watcher instance
:since:  v1.0.0
        """

        vfs_watcher_class = WatcherImplementation.get_class(scheme)
        return vfs_watcher_class()
    #

    @staticmethod
    def get_scheme_from_vfs_url(vfs_url):
        """
Returns the scheme of the VFS URL given.

:param vfs_url: VFS URL to extract the scheme from.

:return: (str) VFS URL scheme
:since:  v1.0.0
        """

        vfs_url = Binary.str(vfs_url)
        if (type(vfs_url) is not str): raise ValueException("VFS URL given is invalid")

        vfs_url_data = vfs_url.split("://", 1)
        if (len(vfs_url_data) == 1): raise ValueException("VFS URL '{0}' is invalid".format(vfs_url))

        return vfs_url_data[0]
    #

    @staticmethod
    def get_scheme_from_vfs_url_if_supported(vfs_url):
        """
Returns the scheme of the VFS URL given if it is supported.

:param vfs_url: VFS URL to extract the scheme from.

:return: (str) VFS URL scheme if supported; None otherwise
:since:  v1.0.0
        """

        _return = None

        if (not WatcherImplementation.is_disabled()):
            try:
                scheme = WatcherImplementation.get_scheme_from_vfs_url(vfs_url)

                if (NamedClassLoader.is_defined_in_namespace("dpt_vfs", "{0}.Watcher".format(scheme.replace("-", "_")))):
                    _return = scheme
                #
            except ValueException: pass
        #

        return _return
    #

    @staticmethod
    def is_disabled():
        """
Returns true VFS watchers are disabled.

:return: (bool) True if disabled
:since:  v1.0.0
        """

        with WatcherImplementation._lock: return WatcherImplementation._is_disabled
    #
#



# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.4
dpt_vfs/dpt_vfs/file/watcher.py
"""

# pylint: disable=import-error,invalid-name,no-name-in-module

try: from urllib.parse import unquote_plus, urlsplit
except ImportError:
    from urllib import unquote_plus
    from urlparse import urlsplit
#

from dpt_logging import LogLine
from dpt_threading import InstanceLock

from ...abstract_watcher import AbstractWatcher
from .watcher_mtime import WatcherMtime

_IMPLEMENTATION_INOTIFY = 1
"""
pyinotify implementation
"""
_IMPLEMENTATION_INOTIFY_SYNC = 2
"""
Synchronous pyinotify implementation
"""
_IMPLEMENTATION_MTIME = 3
"""
Filesystem mtime implementation
"""

try:
    from .watcher_pyinotify import WatcherPyinotify
    from .watcher_pyinotify_sync import WatcherPyinotifySync
    _mode = _IMPLEMENTATION_INOTIFY
except ImportError:
    _mode = _IMPLEMENTATION_MTIME
    WatcherPyinotify = None
#

class Watcher(AbstractWatcher):
    """
"file:///" watcher for change events.

:author:     direct Netware Group et al.
:copyright:  direct Netware Group - All rights reserved
:package:    dpt
:subpackage: vfs
:since:      v1.0.0
:license:    https://www.direct-netware.de/redirect?licenses;mpl2
             Mozilla Public License, v. 2.0
    """

    IMPLEMENTATION_INOTIFY = _IMPLEMENTATION_INOTIFY
    """
pyinotify implementation
    """
    IMPLEMENTATION_INOTIFY_SYNC = _IMPLEMENTATION_INOTIFY_SYNC
    """
Synchronous pyinotify implementation
    """
    IMPLEMENTATION_MTIME = _IMPLEMENTATION_MTIME
    """
Filesystem mtime implementation
    """

    __slots__ = [ ]
    """
python.org: __slots__ reserves space for the declared variables and prevents
the automatic creation of __dict__ and __weakref__ for each instance.
    """
    _implementation = None
    """
Watcher implementation instance
    """
    _instance = None
    """
Watcher implementation instance
    """
    _lock = InstanceLock()
    """
Thread safety lock
    """

    def __init__(self):
        """
Constructor __init__(Watcher)

:since: v1.0.0
        """

        if (Watcher._instance is None):
            with Watcher._lock:
                if (Watcher._instance is None): self.set_implementation()
            #
        #
    #

    @property
    def implementing_scheme(self):
        """
Returns the implementing scheme name.

:return: (str) Implementing scheme name
:since:  v1.0.0
        """

        return "file"
    #

    @property
    def is_synchronous(self):
        """
Returns true if changes are only detected after "check()" has been
called.

:return: (bool) True if changes are not detected automatically
:since:  v1.0.0
        """

        with Watcher._lock: return (WatcherPyinotify is None or (not isinstance(Watcher._instance, WatcherPyinotify)))
    #

    def check(self, url):
        """
Checks a given URL for changes if "is_synchronous()" is true.

:param url: Resource URL

:since: v1.0.0
        """

        _path = Watcher._get_path(url)

        with Watcher._lock:
            if (Watcher._instance is not None
                and _path is not None
                and _path.strip() != ""
               ): Watcher._instance.check(_path)
        #
    #

    def free(self):
        """
Frees all watcher callbacks for garbage collection.

:since: v1.0.0
        """

        with Watcher._lock:
            if (Watcher._instance is not None): Watcher._instance.free()
        #
    #

    def is_watched(self, url, callback = None):
        """
Returns true if the resource URL is already watched. It will return false
if a callback is given but not defined for the watched URL.

:param url: Resource URL
:param callback: Callback to be checked for the watched resource URL

:return: (bool) True if watched with the defined callback or any if not
         defined.
:since:  v1.0.0
        """

        _path = Watcher._get_path(url)

        with Watcher._lock:
            return (False
                    if (Watcher._instance is None or _path is None or _path.strip() == "") else
                    Watcher._instance.is_watched(_path, callback)
                   )
        #
    #

    def register(self, url, callback):
        """
Handles registration of resource URL watches and its callbacks.

:param url: Resource URL to be watched

:return: (bool) True on success
:since:  v1.0.0
        """

        _path = Watcher._get_path(url)

        with Watcher._lock:
            return (False
                    if (Watcher._instance is None or _path is None or _path.strip() == "") else
                    Watcher._instance.register(_path, callback)
                   )
        #
    #

    def set_implementation(self, implementation = None):
        """
Set the filesystem watcher implementation to use.

:param implementation: Implementation identifier

:since: v1.0.0
        """

        # global: _IMPLEMENTATION_INOTIFY, _IMPLEMENTATION_INOTIFY_SYNC, _IMPLEMENTATION_MTIME, _mode

        with Watcher._lock:
            if (Watcher._instance is not None): Watcher._instance.stop()

            if (_mode == _IMPLEMENTATION_INOTIFY
                and (implementation is None or implementation == _IMPLEMENTATION_INOTIFY)
            ): Watcher._implementation = _IMPLEMENTATION_INOTIFY
            elif (_mode == _IMPLEMENTATION_INOTIFY
                and implementation == _IMPLEMENTATION_INOTIFY_SYNC
                ): Watcher._implementation = _IMPLEMENTATION_INOTIFY_SYNC
            else: Watcher._implementation = _IMPLEMENTATION_MTIME

            Watcher._init_watcher()
        #
    #

    def stop(self):
        """
Stops all watchers.

:since: v1.0.0
        """

        with Watcher._lock:
            if (Watcher._instance is not None): Watcher._instance.stop()
        #
    #

    def unregister(self, url, callback):
        """
Handles deregistration of resource URL watches.

:param url: Resource URL watched

:return: (bool) True on success
:since:  v1.0.0
        """

        _path = Watcher._get_path(url)

        with Watcher._lock:
            return (False
                    if (Watcher._instance is None or _path is None or _path.strip() == "") else
                    Watcher._instance.unregister(_path, callback)
                   )
        #
    #

    @staticmethod
    def disable():
        """
Disables this watcher and frees all callbacks for garbage collection.

:since: v1.0.0
        """

        with Watcher._lock:
            if (Watcher._instance is not None):
                Watcher._instance.stop()
                Watcher._instance = None

                LogLine.debug("dpt_vfs.file.Watcher has been disabled", context = "dpt_vfs")
            #
        #
    #

    @staticmethod
    def _get_path(url):
        """
Return the local filesystem path for the given "file:///" URL.

:param url: Filesystem URL

:return: (str) Filesystem path; None if not a "file:///" URL
:since:  v1.0.0
        """

        url_elements = urlsplit(url)
        return (unquote_plus(url_elements.path[1:]) if (url_elements.scheme == "file") else None)
    #

    @staticmethod
    def _init_watcher():
        """
Initializes the watcher instance.

:since: v1.0.0
        """

        instance_callable = WatcherMtime

        if (WatcherPyinotify is not None):
            if (Watcher._implementation == _IMPLEMENTATION_INOTIFY): instance_callable = WatcherPyinotify.get_singleton
            elif (Watcher._implementation == _IMPLEMENTATION_INOTIFY_SYNC): instance_callable = WatcherPyinotifySync.get_singleton
        #

        try: Watcher._instance = instance_callable()
        except OSError:
            if (instance_callable is WatcherMtime): raise
            Watcher._instance = WatcherMtime()
        #
    #
#



# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.4
dpt_vfs/dpt_vfs/file/watcher_pyinotify_callback.py
"""

# pylint: disable=import-error,no-name-in-module

from os import path
from weakref import ref

try: from urllib.parse import quote_plus
except ImportError: from urllib import quote_plus

from dpt_logging import ExceptionLogTrap
from dpt_runtime import Binary
from pyinotify import ProcessEvent

from ...abstract_watcher import AbstractWatcher

class WatcherPyinotifyCallback(ProcessEvent):
    """
Processes pyinotify events and calls defined callbacks.

:author:     direct Netware Group et al.
:copyright:  direct Netware Group - All rights reserved
:package:    dpt
:subpackage: vfs
:since:      v1.0.0
:license:    https://www.direct-netware.de/redirect?licenses;mpl2
             Mozilla Public License, v. 2.0
    """

    # pylint: disable=invalid-name

    # pylint: disable=bad-option-value,slots-on-old-class
    __slots__ = [ "manager_weakref" ]
    """
python.org: __slots__ reserves space for the declared variables and prevents
the automatic creation of __dict__ and __weakref__ for each instance.
    """

    def __init__(self, manager):
        """
Constructor __init__(WatcherPyinotifyCallback)

:since: v1.0.0
        """

        ProcessEvent.__init__(self)

        self.manager_weakref = ref(manager)
        """
pyinotify manager instance
        """
    #

    def _process_callbacks(self, event_type, _path, changed_value = None):
        """
Handles all inotify events.

:param event_type: Event type defined in AbstractWatcher
:param _path: Filesystem path
:param changed_value: Changed value (e.g. name of deleted or created file)

:since: v1.0.0
        """

        changed_value = Binary.str(changed_value)
        manager = self.manager_weakref()
        _path = Binary.str(_path)

        changed_path = (_path if (changed_value is None) else path.join(_path, changed_value))

        if (manager and manager.is_watched(changed_path)):
            callbacks = manager.get_callbacks(changed_path)
            url = "file:///{0}".format(quote_plus(_path, "/"))

            for callback in callbacks:
                with ExceptionLogTrap("dpt_vfs"): callback(event_type, url, changed_value)
            #
        #
    #

    def process_IN_ATTRIB(self, event):
        """
Handles "IN_ATTRIB" inotify events.

:param event: pyinotify event

:since: v1.0.0
        """

        self._process_callbacks(AbstractWatcher.EVENT_TYPE_MODIFIED, event.pathname)
    #

    def process_IN_CLOSE_WRITE(self, event):
        """
Handles "IN_CLOSE_WRITE" inotify events.

:param event: pyinotify event

:since: v1.0.0
        """

        self._process_callbacks(AbstractWatcher.EVENT_TYPE_MODIFIED, event.pathname)
    #

    def process_IN_CREATE(self, event):
        """
Handles "IN_CREATE" inotify events.

:param event: pyinotify event

:since: v1.0.0
        """

        dir_path = path.dirname(event.pathname)
        name = path.basename(event.pathname)

        self._process_callbacks(AbstractWatcher.EVENT_TYPE_CREATED, dir_path, name)
    #

    def process_IN_DELETE(self, event):
        """
Handles "IN_DELETE" inotify events.

:param event: pyinotify event

:since: v1.0.0
        """

        dir_path = path.dirname(event.pathname)
        name = path.basename(event.pathname)

        self._process_callbacks(AbstractWatcher.EVENT_TYPE_DELETED, dir_path, name)
    #

    def process_IN_DELETE_SELF(self, event):
        """
Handles "IN_DELETE_SELF" inotify events.

:param event: pyinotify event

:since: v1.0.0
        """

        manager = self.manager_weakref()
        if (manager): manager.unregister(event.pathname, None, True)

        self._process_callbacks(AbstractWatcher.EVENT_TYPE_DELETED, event.pathname)
    #

    def process_IN_MOVE_SELF(self, event):
        """
Handles "IN_MOVE_SELF" inotify events.

:param event: pyinotify event

:since: v1.0.0
        """

        manager = self.manager_weakref()
        if (manager): manager.unregister(event.pathname, None, True)

        self._process_callbacks(AbstractWatcher.EVENT_TYPE_DELETED, event.pathname)
    #

    def process_IN_MOVED_FROM(self, event):
        """
Handles "IN_MOVED_FROM" inotify events.

:param event: pyinotify event

:since: v1.0.0
        """

        dir_path = path.dirname(event.pathname)
        name = path.basename(event.pathname)

        self._process_callbacks(AbstractWatcher.EVENT_TYPE_DELETED, dir_path, name)
    #

    def process_IN_MOVED_TO(self, event):
        """
Handles "IN_MOVED_TO" inotify events.

:param event: pyinotify event

:since: v1.0.0
        """

        dir_path = path.dirname(event.pathname)
        name = path.basename(event.pathname)

        self._process_callbacks(AbstractWatcher.EVENT_TYPE_CREATED, dir_path, name)
    #
#
